
#include "include/Common.fx"

//=======================================================================================
// Macros
//=======================================================================================

#if NBROFSPLITS > 1
#define NUMSPLITS NBROFSPLITS
#else
#define NUMSPLITS 1
#endif

//=======================================================================================
// Structures
//=======================================================================================

struct VSDepthMapIn
{
    float3 pos  : POSITION;         // position
    float2 tex  : TEXCOORD;         // texture coordinate
};

struct VSDepthMapOut
{
    float4 pos   : SV_POSITION;     // position
    float2 tex   : TEXCOORD0;       // texture coordinate
    float depth  : Depth;			// depth
};

struct GSDepthMapOut
{
    float4 pos   : SV_POSITION;				// position
    float2 tex   : TEXCOORD0;				// texture coordinate
    float depth  : Depth;
    uint vpIndex : SV_ViewportArrayIndex;	// viewport index
};

struct PSDepthMapIn
{
    float4 pos   : SV_POSITION;     // position
    float2 tex   : TEXCOORD0;       // texture coordinate
    float depth  : Depth;
};

//=======================================================================================
// Uniforms
//=======================================================================================

cbuffer cbPerFrame
{
	float4x4 g_mView;
    float4x4 g_mProj;
    float4x4 g_mLightViewProj;
    float4x4 g_mCrop[NUMSPLITS];
    float4x4 g_mLiSP[NUMSPLITS];
};

cbuffer cbPerBatch
{
	int g_firstSplit;
	int g_lastSplit;
};

//=======================================================================================
// Vertex shaders
//=======================================================================================

VSDepthMapOut VSSMmain(VSDepthMapIn input)
{
    VSDepthMapOut output = (VSDepthMapOut)0;

    float4 worldPos  = mul( float4( input.pos, 1 ), g_mWorld );
    
    output.pos = mul( worldPos, g_mLightViewProj );
    output.tex = mul( float4( input.tex, 1, 1 ), g_mDiffuseTex ).xy;
    
	return output;
}

VSDepthMapOut VSPreDepth(VSDepthMapIn input)
{
	VSDepthMapOut output = (VSDepthMapOut)0;
	
	float4 worldPos  = mul( float4( input.pos, 1 ), g_mWorld );
    float4 camPos = mul( worldPos, g_mView );
    
    output.pos = mul( camPos, g_mProj );	
	output.depth = -camPos.z;
	
	return output;
}

//=======================================================================================
// Pixel shaders
//=======================================================================================

float PSHWSMmain(PSDepthMapIn input) : SV_Target
{   
    float alpha = matDiffuse.a;
    
    if ( textured )
		alpha = texDiffuse.Sample(SampPointWrap, input.tex).a;
	
	if ( alpha < 0.5 )
		discard;
	
	return 1.0f;
}

float4 PSVSMmain(PSDepthMapIn input) : SV_Target
{   
    float alpha = matDiffuse.a;
    
    if ( textured )
		alpha = texDiffuse.Sample(SampPointWrap, input.tex).a;
	
	if ( alpha < 0.5 )
		discard;

	float posDepth = exp(ExpC * input.pos.z);
	float negDepth = -exp(-ExpC * input.pos.z);

	return float4(posDepth, posDepth * posDepth, negDepth, negDepth * negDepth);
}

float2 PSPreDepth(PSDepthMapIn input) : SV_Target
{
	return (float2)input.depth;
}

//=======================================================================================
// Geometry shaders
//=======================================================================================

[maxvertexcount(NUMSPLITS * 3)]
void GSSMmain(triangle VSDepthMapOut input[3], inout TriangleStream<GSDepthMapOut> stream)
{
	// for all splits which contain the current object
	for (int split = g_firstSplit; split <= g_lastSplit; split++)
	{
		GSDepthMapOut output = (GSDepthMapOut)0;
		
		// set the viewport index corresponding to the correct area in the texture atlas
		// according to the current split
		output.vpIndex = split;

		[unroll] for (int vertex = 0; vertex < 3; vertex++)
		{
			output.pos = mul( input[vertex].pos, g_mLiSP[split] );
			output.pos = mul( output.pos, g_mCrop[split] );
			output.tex = input[vertex].tex;
			stream.Append(output);
		}
		stream.RestartStrip();
	}
}

//=======================================================================================
// Techniques
//=======================================================================================

technique10 RenderHWSMDepthMap
{
    pass p0
    {
        SetVertexShader( CompileShader( vs_4_0, VSSMmain() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSHWSMmain() ) );
        
        SetDepthStencilState( DSEnableDepthLessEqual, 0x00000000 );
        SetBlendState( BSNoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetRasterizerState( RSLightRenderDepth );
    }
    
    pass p1
    {
        SetVertexShader( CompileShader( vs_4_0, VSSMmain() ) );
        SetGeometryShader( CompileShader(gs_4_0, GSSMmain()) );
        SetPixelShader( CompileShader( ps_4_0, PSHWSMmain() ) );
        
        SetDepthStencilState( DSEnableDepthLessEqual, 0x00000000 );
        SetBlendState( BSNoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetRasterizerState( RSLightRenderDepth );
    }
}

technique10 RenderVSMDepthMap
{
    pass p0
    {
        SetVertexShader( CompileShader( vs_4_0, VSSMmain() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSVSMmain() ) );
        
        SetDepthStencilState( DSEnableDepthLessEqual, 0x00000000 );
        SetBlendState( BSNoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetRasterizerState( RSSolidFillCullBack );
    }
    
    pass p1
	{
		SetVertexShader( CompileShader( vs_4_0, VSSMmain() ) );
		SetGeometryShader( CompileShader(gs_4_0, GSSMmain()) );
		SetPixelShader( CompileShader( ps_4_0, PSVSMmain() ) );
	    
		SetDepthStencilState( DSEnableDepthLessEqual, 0x00000000 );
		SetBlendState( BSNoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetRasterizerState( RSSolidFillCullBack );
	}
}

technique10 RenderPreDepth
{
	pass p0
	{
		SetVertexShader( CompileShader( vs_4_0, VSPreDepth() ) );
        SetGeometryShader( NULL );
        SetPixelShader( CompileShader( ps_4_0, PSPreDepth() ) );
        
        SetDepthStencilState( DSEnableDepthLessEqual, 0x00000000 );
        SetBlendState( BSNoBlending, float4( 0.0f, 0.0f, 0.0f, 0.0f ), 0xFFFFFFFF );
        SetRasterizerState( RSSolidFillCullBack );		
	}
}



