#include "include/Poisson.fx"
#include "include/Common.fx"

//=============================================================================
// Constants
//=============================================================================

static const float LBRAmount = 0.5f;
static const float MinVariance = 0.000001f;

//=============================================================================
// Macros
//=============================================================================

// disable warning 3570: gradient instruction in a loop with varying iteration
#pragma warning( disable : 3570 )

#ifdef PRESET
#define PCF_PRESET PRESET
#else
#define PCF_PRESET 2
#endif

#if PCF_PRESET == 0

#define SEARCH_POISSON_COUNT 16		// 16x Poisson Blockersearch
#define SEARCH_POISSON Poisson16
#define PCF_POISSON_COUNT 25		// 25x Poisson PCF
#define PCF_POISSON Poisson25

#define SEARCH_BOX_COUNT 1			// 2x2 Box Blockersearch
#define PCF_BOX_COUNT 2				// 4x4 Box PCF

#elif PCF_PRESET == 1

#define SEARCH_POISSON_COUNT 25		// 25x Poisson Blockersearch
#define SEARCH_POISSON Poisson25
#define PCF_POISSON_COUNT 32		// 32x Poisson PCF
#define PCF_POISSON Poisson32

#define SEARCH_BOX_COUNT 2			// 4x4 Box Blockersearch
#define PCF_BOX_COUNT 3				// 6x6 Box PCF

#else

#define SEARCH_POISSON_COUNT 32		// 32x Poisson Blockersearch
#define SEARCH_POISSON Poisson32
#define PCF_POISSON_COUNT 64		// 64x Poisson PCF
#define PCF_POISSON Poisson64

#define SEARCH_BOX_COUNT 3			// 6x6 Box Blockersearch
#define PCF_BOX_COUNT 4				// 8x8 Box PCF

#endif

//=============================================================================
// Uniforms
//=============================================================================

float g_LightZNear = 0.1;
float g_LightZFar = 600.0;

//=============================================================================
// Functions
//=============================================================================

float3 ShadowLookUp(Texture2D shadowMap, float2 uv, float zReceiver)
{
	return (float3)shadowMap.SampleCmpLevelZero(SampCmp, uv, zReceiver);
}

float3 HWPCFShadowLookUp(Texture2D shadowMap, float2 uv, float zReceiver)
{
	return (float3)shadowMap.SampleCmpLevelZero(SampPCF, uv, zReceiver);
}

float3 ShadowLookUpArray(Texture2DArray shadowMap, float2 uv, int index, float zReceiver)
{
	float depth = shadowMap.Sample(SampPointBorder, float3(uv.x, uv.y, index)).r;
	
	return depth < zReceiver ? (float3)0.0f : (float3)1.0f;
}

// Light bleeding reduction
float ReduceLightBleeding(float p)
{
    return smoothstep(LBRAmount, 1.0f, p);
}

float ChebyshevUpperBound(float2 moments, float mean, float minVariance)
{
    // Standard shadow map comparison
    float p = (mean <= moments.x);
    
    // Compute variance
    float variance = moments.y - (moments.x * moments.x);
    variance = max(variance, minVariance);
    
    // Compute probabilistic upper bound
    float d = mean - moments.x;
    float p_max = variance / (variance + d * d);
    
    return max(p, p_max);
}

float3 ExpVSMLookUp(Texture2D shadowMap, float2 uv, float zReceiver, float minVariance)
{
	float4 moments = shadowMap.Sample(SampVSM, uv);
    float2 posMoments = moments.xy;
    float2 negMoments = moments.zw;
    float expPosDepth = exp(ExpC * zReceiver);
	float expNegDepth = -exp(-ExpC * zReceiver);
    
    float posDepthScale = ExpC * expPosDepth;
    float posMinVariance = minVariance * (posDepthScale * posDepthScale);
    float posShadowContrib = ChebyshevUpperBound(posMoments, expPosDepth, posMinVariance);
    
    float negDepthScale = ExpC * expNegDepth;
    float negMinVariance = minVariance * (negDepthScale * negDepthScale);
    float negShadowContrib = ChebyshevUpperBound(negMoments, expNegDepth, negMinVariance);
    
    float shadowContrib = min(posShadowContrib, negShadowContrib);
	
	shadowContrib = ReduceLightBleeding(shadowContrib);
	
    //return clamp(ShadowContrib, 0.4, 1.0);
    return (float3)shadowContrib;
}

// Using similar triangles from the surface point to the area light
float2 SearchRegionRadiusUV(float2 lightRadius, float zWorld)
{
    return lightRadius * (zWorld - g_LightZNear) / zWorld;
}

// Using similar triangles between the area light, the blocking plane and the surface point
float2 PenumbraRadiusUV(float2 lightRadius, float zReceiver, float zBlocker)
{
    return lightRadius * (zReceiver - zBlocker) / zBlocker;
}

// Project UV size to the near plane of the light
float2 ProjectToLightUV(float2 sizeUV, float zWorld)
{
    return sizeUV * g_LightZNear / zWorld;
}

// Derivatives of light-space depth with respect to texture coordinates
float2 DepthGradient(float2 uv, float z)
{
    float2 dz_duv = 0;

    float3 duvdist_dx = ddx(float3(uv, z));
    float3 duvdist_dy = ddy(float3(uv, z));

    dz_duv.x = duvdist_dy.y * duvdist_dx.z;
    dz_duv.x -= duvdist_dx.y * duvdist_dy.z;
    
    dz_duv.y = duvdist_dx.x * duvdist_dy.z;
    dz_duv.y -= duvdist_dy.x * duvdist_dx.z;

    float det = (duvdist_dx.x * duvdist_dy.y) - (duvdist_dx.y * duvdist_dy.x);
    dz_duv /= det;

    return dz_duv;
}

float BiasedZ(float z, float2 dz_duv, float2 offset)
{
    return z + dot(dz_duv, offset);
}

float ZClipToZEye(float zClip)
{
	return g_LightZFar * g_LightZNear / (g_LightZFar - zClip * (g_LightZFar - g_LightZNear));   
}

float3 PoissonPCF(Texture2D shadowMap, float2 uv, float zReceiver, float texMapSize, bool hwPCF)
{
    float sum = 0;
    float2 texelSz = float2(1.0 / texMapSize, 1.0 / texMapSize);
    
    for (int i = 0; i < PCF_POISSON_COUNT; ++i)
    {
        float2 offset = PCF_POISSON[i] * texelSz;
        float z = zReceiver;
        
        if (hwPCF) sum += HWPCFShadowLookUp(shadowMap, uv + offset, z).r;
		else sum += ShadowLookUp(shadowMap, uv + offset, z).r;
    }
    
    return (float3)sum / PCF_POISSON_COUNT;
}

float3 PoissonPCFShadowLookUp(Texture2D shadowMap, float2 uv, float zReceiver, float2 dz_duv, float2 filterRadiusUV, bool hwPCF)
{
    float sum = 0;
    
    for (int i = 0; i < PCF_POISSON_COUNT; ++i)
    {
        float2 offset = PCF_POISSON[i] * filterRadiusUV;
        float z = BiasedZ(zReceiver, dz_duv, offset);
        
        if (hwPCF) sum += HWPCFShadowLookUp(shadowMap, uv + offset, z).r;
		else sum += ShadowLookUp(shadowMap, uv + offset, z).r;
    }
    
    return (float3)sum / PCF_POISSON_COUNT;
}

float3 BoxPCF(Texture2D shadowMap, float2 uv, float zReceiver, float texMapSize, bool hwPCF)
{
    float sum = 0;
    float2 texelSz = float2(1.0 / texMapSize, 1.0 / texMapSize);
    
	for (float x = -PCF_BOX_COUNT + 0.5; x <= PCF_BOX_COUNT - 0.5; ++x)
		for (float y = -PCF_BOX_COUNT + 0.5; y <= PCF_BOX_COUNT - 0.5; ++y)
        {
            float2 offset = float2( x, y ) * texelSz;
            float z = zReceiver;
            
            if (hwPCF) sum += HWPCFShadowLookUp(shadowMap, uv + offset, z).r;
			else sum += ShadowLookUp(shadowMap, uv + offset, z).r;
        }
    float numSamples = (2 * PCF_BOX_COUNT);
    
    return (float3)sum / (numSamples * numSamples);
}

float3 BoxPCFShadowLookUp(Texture2D shadowMap, float2 uv, float zReceiver, float2 dz_duv, float2 filterRadiusUV, bool hwPCF)
{
    float sum = 0;
    
	float2 stepUV = filterRadiusUV / PCF_BOX_COUNT;
    for (float x = -PCF_BOX_COUNT + 0.5; x <= PCF_BOX_COUNT - 0.5; ++x)
        for (float y = -PCF_BOX_COUNT + 0.5; y <= PCF_BOX_COUNT - 0.5; ++y)
        {
            float2 offset = float2( x, y ) * stepUV;
            float z = BiasedZ(zReceiver, dz_duv, offset);
            
            if (hwPCF) sum += HWPCFShadowLookUp(shadowMap, uv + offset, z).r;
			else sum += ShadowLookUp(shadowMap, uv + offset, z).r;
        }
    float numSamples = (2 * PCF_BOX_COUNT);
    
    return (float3)sum / (numSamples * numSamples);
}

// Returns average blocker depth in the search region, as well as the number of found blockers.
// Blockers are defined as shadow-map samples between the surface point and the light.
float FindBlockersPoisson(out float avgBlockerDepth, Texture2D shadowMap,
				   float2 uv, float zReceiver, float2 dz_duv, float2 searchRegionRadiusUV)
{
    float blockerDepth = 0;
    float numBlockers = 0;

    for (int i = 0; i < SEARCH_POISSON_COUNT; ++i)
    {
        float2 offset = SEARCH_POISSON[i] * searchRegionRadiusUV;
        float depth = shadowMap.SampleLevel(SampAnisotropicClamp, uv + offset, 0).r;
        float z = BiasedZ(zReceiver, dz_duv, offset);
        if ( depth < z )
        {
            blockerDepth += depth;
            numBlockers++;
        }
    }
    avgBlockerDepth = blockerDepth / numBlockers;
    
    return numBlockers;
}

// Returns average blocker depth in the search region, as well as the number of found blockers.
// Blockers are defined as shadow-map samples between the surface point and the light.
float FindBlockersBox(out float avgBlockerDepth, Texture2D shadowMap,
				   float2 uv, float zReceiver, float2 dz_duv, float2 searchRegionRadiusUV)
{
    float blockerDepth = 0;
    float numBlockers = 0;

    float2 stepUV = searchRegionRadiusUV / SEARCH_BOX_COUNT;
    for (float x = -SEARCH_BOX_COUNT + 0.5; x <= SEARCH_BOX_COUNT - 0.5; ++x)
        for (float y = -SEARCH_BOX_COUNT + 0.5; y <= SEARCH_BOX_COUNT - 0.5; ++y)
        {
            float2 offset = float2( x, y ) * stepUV;
            float depth = shadowMap.SampleLevel(SampAnisotropicClamp, uv + offset, 0).r;
            float z = BiasedZ(zReceiver, dz_duv, offset);
            if ( depth < z )
            {
                blockerDepth += depth;
                numBlockers++;
            }
        }

    avgBlockerDepth = blockerDepth / numBlockers;
    
    return numBlockers;
}

float3 PoissonPCFPCSSLookUp(Texture2D shadowMap, float2 uv, float zReceiver, float zEye, float2 lightRadius, bool hwPCF)
{
    // ------------------------
    // blocker search
    // ------------------------
    //float2 dz_duv = DepthGradient(uv, zReceiver);
    float2 dz_duv = 0;
    float2 searchRegionRadiusUV = SearchRegionRadiusUV(lightRadius, zEye);
	float avgBlockerDepth = 0;
    float numBlockers = FindBlockersPoisson(avgBlockerDepth, shadowMap, uv, zReceiver, dz_duv, searchRegionRadiusUV);

    // Early out if no blocker found
    if (numBlockers == 0) return (float3)1.0;

    // ------------------------
    // penumbra size
    // ------------------------
    float avgBlockerDepthWorld = ZClipToZEye(avgBlockerDepth);
    float2 penumbraRadiusUV = PenumbraRadiusUV(lightRadius, zEye, avgBlockerDepthWorld);
    
    // ------------------------
    // filtering
    // ------------------------
    float2 filterRadiusUV = ProjectToLightUV(penumbraRadiusUV, zEye);
    return PoissonPCFShadowLookUp(shadowMap, uv, zReceiver, dz_duv, filterRadiusUV, hwPCF);
}

float3 BoxPCFLookUp(Texture2D shadowMap, float2 uv, float zReceiver, float zEye, float2 lightRadius, bool hwPCF, bool blockerSearch)
{
    //float2 dz_duv = DepthGradient(uv, zReceiver);
    float2 dz_duv = 0;
    float numBlockers = 1;
    
    if (blockerSearch)
    {
		// ------------------------
		// blocker search
		// ------------------------		
		float2 searchRegionRadiusUV = SearchRegionRadiusUV(lightRadius, zEye);
		float avgBlockerDepth = 0;
		numBlockers = FindBlockersBox(avgBlockerDepth, shadowMap, uv, zReceiver, dz_duv, searchRegionRadiusUV);
	}
	
	// Early out if no blocker found
	if (numBlockers == 0) return (float3)1.0;
	
	// ------------------------
    // filtering
    // ------------------------
    float2 filterRadiusUV = lightRadius;
    return BoxPCFShadowLookUp(shadowMap, uv, zReceiver, dz_duv, filterRadiusUV, hwPCF);
}

float3 PoissonPCFLookUp(Texture2D shadowMap, float2 uv, float zReceiver, float zEye, float2 lightRadius, bool hwPCF, bool blockerSearch)
{
	//float2 dz_duv = DepthGradient(uv, zReceiver);
    float2 dz_duv = 0;
	float numBlockers = 1;
    
    if (blockerSearch)
    {
		// ------------------------
		// blocker search
		// ------------------------
		float2 searchRegionRadiusUV = SearchRegionRadiusUV(lightRadius, zEye);
		float avgBlockerDepth = 0;	    
		numBlockers = FindBlockersPoisson(avgBlockerDepth, shadowMap, uv, zReceiver, dz_duv, searchRegionRadiusUV);
	}
	
	// Early out if no blocker found
	if (numBlockers == 0) return (float3)1.0;
	
	// ------------------------
    // filtering
    // ------------------------
    float2 filterRadiusUV = lightRadius;
    return PoissonPCFShadowLookUp(shadowMap, uv, zReceiver, dz_duv, filterRadiusUV, hwPCF);
}
