#include "stdafx.h"

#include <stdio.h>
#include <stdarg.h>

#include "D3dRenderSystem.h"

Logger D3dFontManager::logger("D3dFontManager");

D3dFontManager::D3dFontManager(ID3D10Device& d3dDevice) :
	_d3dDevice(&d3dDevice),
	_fonts(1, 0),
	_d3dFonts(1, 0),
	_d3dSprites(1, 0),
	_fontColor(1.0f, 1.0f, 1.0f, 1.0f),
	_shadowColor(.0f, .0f, .0f, 1.0f),
	_shadowed(false),
	_lineHeight(0)
{
}

D3dFontManager::~D3dFontManager()
{	
	for (unsigned int i = 0; i < _d3dFonts.size(); ++i)
		if (_d3dFonts[i])
			_d3dFonts[i]->Release();

	for (unsigned int i = 0; i < _d3dSprites.size(); ++i)
		if (_d3dSprites[i])
			_d3dSprites[i]->Release();
}

bool D3dFontManager::Initialize(const D3dRenderSystem&)
{
	HRESULT hr = S_OK;

	return hr == S_OK;
}

bool D3dFontManager::Generate(Font& font, const D3dRenderSystem &renderSystem)
{
	bool found = false;
	
	for (unsigned i = 1; i < _fonts.size() && !found; ++i)
	{
		if (_fonts[i]->GetName() == font.GetName() &&
			_fonts[i]->GetSize() == font.GetSize() &&
			_fonts[i]->GetBoldFlag() == font.GetBoldFlag() &&
			_fonts[i]->GetItalicFlag() == font.GetItalicFlag())
		{
			font.SetId(_fonts[i]->GetId());
			found = true;
		}
	}

	if (!found)
	{
		ID3DX10Font* d3dFont = 0;
		ID3DX10Sprite* d3dSprite = 0;
		HRESULT hr = S_OK;

		if ((hr = D3DX10CreateFont(_d3dDevice, font.GetSize(), 0, font.GetBoldFlag() ? FW_BOLD : FW_NORMAL, 1,
								   font.GetItalicFlag() ? TRUE : FALSE, DEFAULT_CHARSET, OUT_DEFAULT_PRECIS,
								   DEFAULT_QUALITY, DEFAULT_PITCH | FF_DONTCARE, font.GetName().c_str(), &d3dFont)) != S_OK)
		{
			if (logger.IsEnabled(Logger::Error))
				Logger::Stream(logger, Logger::Error) << "Cannot create font " << font.GetName() <<
					" (" << renderSystem.GetErrorString(hr) << ").";

			return false;
		}

		const unsigned int deviceBufferSize = 512;

		if ((hr = D3DX10CreateSprite(_d3dDevice, deviceBufferSize, &d3dSprite)) != S_OK)
		{
			if (logger.IsEnabled(Logger::Error))
				Logger::Stream(logger, Logger::Error) << "Cannot create sprite for font " << font.GetName() <<
					" (" << renderSystem.GetErrorString(hr) << ").";
			
			return false;
		}

		font.SetId(_fonts.size());
		_fonts.push_back(&font);
		_d3dFonts.push_back(d3dFont);
		_d3dSprites.push_back(d3dSprite);

		if (logger.IsEnabled(Logger::Info))
			Logger::Stream(logger, Logger::Info) << "Generate font " << font.GetName();
	}

	return true;
}

void D3dFontManager::Begin(Font& font)
{
	D3D10_VIEWPORT VPs[D3D10_VIEWPORT_AND_SCISSORRECT_OBJECT_COUNT_PER_PIPELINE];
	UINT cVPs = 1;

	_d3dDevice->RSGetViewports( &cVPs, VPs );
	D3DXMATRIXA16 matProjection;
	D3DXMatrixOrthoOffCenterLH( &matProjection, ( FLOAT )VPs[0].TopLeftX, ( FLOAT )
								( VPs[0].TopLeftX + VPs[0].Width ), ( FLOAT )VPs[0].TopLeftY, ( FLOAT )
								( VPs[0].TopLeftY + VPs[0].Height ), 0.1f, 10.0f );

	_d3dSprites[font.GetId()]->SetProjectionTransform( &matProjection );

	_d3dSprites[font.GetId()]->Begin( /*D3DX10_SPRITE_SORT_TEXTURE |*/ D3DX10_SPRITE_SAVE_STATE );
}

void D3dFontManager::End(Font& font)
{
	_d3dSprites[font.GetId()]->End();
}

void D3dFontManager::SetFontColor(const tVec4f& color)
{
	_fontColor = D3DXCOLOR(color.x, color.y, color.z, color.w);
}

void D3dFontManager::SetShadowColor(const tVec4f& color)
{
	_shadowColor = D3DXCOLOR(color.x, color.y, color.z, color.w);
}

void D3dFontManager::SetInsertionPos(const tVec2i& pos)
{
	_pos = pos;
}

void D3dFontManager::SetShadowed(bool enable)
{
	_shadowed = enable;
}

void D3dFontManager::SetLineHeight(unsigned int lineHeight)
{
	_lineHeight = static_cast<int>(lineHeight);
}

void D3dFontManager::DrawFormattedTextLine(Font& font, const char* text, ...)
{
	const unsigned int bufferSize = 512;
	char strBuffer[bufferSize];

	va_list args;
	va_start( args, text );
	vsprintf_s( strBuffer, bufferSize, text, args );
	strBuffer[511] = '\0';
	va_end( args );

	DrawTextLine(font, strBuffer);
}

void D3dFontManager::DrawTextLine(Font& font, const char* text)
{
	if (_shadowed)
		DrawTextLine(font, text, _pos + 1, _shadowColor);

	DrawTextLine(font, text, _pos, _fontColor);

	_pos.y += _lineHeight;
}

void D3dFontManager::DrawTextLine(Font& font, const char* text, const tVec2i& pos, const D3DXCOLOR& color)
{
	RECT rc;
	SetRect(&rc, pos.x, pos.y, 0, 0);

	_d3dFonts[font.GetId()]->DrawText(_d3dSprites[font.GetId()], text, -1, &rc, DT_NOCLIP, color);
}
