#include "stdafx.h"

#include "DXUTApp.h"

#include <io.h>
#include <fcntl.h>
#include <fstream>

extern bool CALLBACK IsD3DDeviceAcceptable(UINT, UINT, D3D10_DRIVER_TYPE, DXGI_FORMAT, bool, void*);
extern HRESULT CALLBACK OnD3DCreateDevice(ID3D10Device* d3dDevice, const DXGI_SURFACE_DESC* bufferSurfaceDesc, void*);
extern HRESULT CALLBACK OnD3DResizedSwapChain(ID3D10Device* d3dDevice, IDXGISwapChain* swapChain, const DXGI_SURFACE_DESC* bufferSurfaceDesc, void*);
extern void CALLBACK OnD3DFrameRender(ID3D10Device*, double, float elapsedTime, void*);
extern LRESULT CALLBACK MsgProc(HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam, bool* noFurtherProcessing, void*);
extern void CALLBACK OnKeyboard(UINT key, bool keyDown, bool altDown, void*);
extern void CALLBACK OnMouse(bool leftDown, bool rightDown, bool middleDown, bool, bool, int wheelDelta, int x, int y, void*);
extern void CALLBACK OnFrameMove(double, float elapsedTime, void*);
extern void CALLBACK OnD3DReleasingSwapChain(void*);
extern void CALLBACK OnD3DDestroyDevice(void*);
extern void CALLBACK OnD3DGUIEvent(UINT eventId, int controlId, CDXUTControl* control, void*);

Logger DXUTApp::logger("DXUTApp");

DXUTApp::DXUTApp() :
	_width(1024),
	_height(768),
	_fullscreen(false),
	_config_fov(60),
	_config_zNear(1),
	_config_zFar(1500),
	_swapChain(0),
	_init(false)
{
}

DXUTApp::~DXUTApp()
{
}

bool DXUTApp::Init()
{
	DXUTSetCallbackMsgProc( MsgProc );
	DXUTSetCallbackFrameMove( OnFrameMove );
	//DXUTSetCallbackKeyboard( OnKeyboard );
	//DXUTSetCallbackMouse( OnMouse, true );
	DXUTSetCallbackD3D10DeviceAcceptable(IsD3DDeviceAcceptable);
	DXUTSetCallbackD3D10DeviceCreated( OnD3DCreateDevice );
	DXUTSetCallbackD3D10SwapChainResized( OnD3DResizedSwapChain );
	DXUTSetCallbackD3D10FrameRender( OnD3DFrameRender );
	DXUTSetCallbackD3D10SwapChainReleasing( OnD3DReleasingSwapChain );
	DXUTSetCallbackD3D10DeviceDestroyed( OnD3DDestroyDevice );

	_settingsDlg.Init(&_dialogResourceManager);

	HRESULT hr = S_OK;

	if ((hr = DXUTInit(true, true, 0, false)) != S_OK)
	{
		PrintErrorMsg(logger, GetDXUTErrorMsg(hr), GetDXUTErrorMsg(hr), L"Init Error");
		return false;
	}
	
	DXUTSetCursorSettings(true, true);

	return true;
}

bool DXUTApp::Create(const std::wstring& title, HINSTANCE instance, int x, int y)
{
	HRESULT hr = S_OK;

	if ((hr = DXUTCreateWindow(title.c_str(), instance, 0, 0, x, y)) != S_OK)
	{
		PrintErrorMsg(logger, GetDXUTErrorMsg(hr), GetDXUTErrorMsg(hr), L"Window Creation Error");
		return false;
	}

	// Default settings
	DXUTDeviceSettings ds;
	ZeroMemory(&ds, sizeof(ds));
	ds.ver                     = DXUT_D3D10_DEVICE;
	ds.d3d10.DriverType        = D3D10_DRIVER_TYPE_HARDWARE;
	ds.d3d10.CreateFlags       = 0;
#if defined(DEBUG) || defined(_DEBUG)
	ds.d3d10.CreateFlags      |= D3D10_CREATE_DEVICE_DEBUG;
#endif
	ds.d3d10.SyncInterval		    = 0;
	ds.d3d10.AutoCreateDepthStencil = true;
	ds.d3d10.AutoDepthStencilFormat = DXGI_FORMAT_D24_UNORM_S8_UINT;

	// Back buffer
	ds.d3d10.sd.BufferDesc.Width			= _width;
	ds.d3d10.sd.BufferDesc.Height			= _height;
	ds.d3d10.sd.BufferDesc.Format			= DXGI_FORMAT_R8G8B8A8_UNORM_SRGB;
	ds.d3d10.sd.BufferDesc.ScanlineOrdering = DXGI_MODE_SCANLINE_ORDER_UNSPECIFIED;
	ds.d3d10.sd.BufferDesc.Scaling			= DXGI_MODE_SCALING_UNSPECIFIED;
	ds.d3d10.sd.BufferUsage					= 0;
	ds.d3d10.sd.OutputWindow				= 0;
	ds.d3d10.sd.Windowed					= true;
	ds.d3d10.sd.SwapEffect					= DXGI_SWAP_EFFECT_DISCARD;
	ds.d3d10.sd.Flags						= DXGI_SWAP_CHAIN_FLAG_ALLOW_MODE_SWITCH;
	// Multisampling
	ds.d3d10.sd.SampleDesc.Count   = 4;
	ds.d3d10.sd.SampleDesc.Quality = 4;

	// Fill in the rest with DXUT defaults
	DXUTMatchOptions mo;
	mo.eAPIVersion       = DXUTMT_PRESERVE_INPUT;
	mo.eAdapterOrdinal   = DXUTMT_IGNORE_INPUT;
	mo.eOutput           = DXUTMT_IGNORE_INPUT;
	mo.eDeviceType       = DXUTMT_PRESERVE_INPUT;
	mo.eWindowed         = DXUTMT_CLOSEST_TO_INPUT;
	mo.eAdapterFormat    = DXUTMT_CLOSEST_TO_INPUT;
	mo.eVertexProcessing = DXUTMT_CLOSEST_TO_INPUT;
	mo.eResolution       = DXUTMT_CLOSEST_TO_INPUT;
	mo.eBackBufferFormat = DXUTMT_CLOSEST_TO_INPUT;
	mo.eBackBufferCount  = DXUTMT_CLOSEST_TO_INPUT;
	mo.eMultiSample      = DXUTMT_CLOSEST_TO_INPUT;
	mo.eSwapEffect       = DXUTMT_PRESERVE_INPUT;
	mo.eDepthFormat      = DXUTMT_CLOSEST_TO_INPUT;
	mo.eStencilFormat    = DXUTMT_IGNORE_INPUT;
	mo.ePresentFlags     = DXUTMT_CLOSEST_TO_INPUT;
	mo.eRefreshRate      = DXUTMT_IGNORE_INPUT;
	mo.ePresentInterval  = DXUTMT_CLOSEST_TO_INPUT;

	if ((hr = DXUTFindValidDeviceSettings(&ds, &ds, &mo)) != S_OK)
	{
		PrintErrorMsg(logger, GetDXUTErrorMsg(hr), GetDXUTErrorMsg(hr), L"Device Creation Error");
		return false;
	}

	if ((hr = DXUTCreateDeviceFromSettings(&ds)) != S_OK)
	{
		PrintErrorMsg(logger, GetDXUTErrorMsg(hr), GetDXUTErrorMsg(hr), L"Device Creation Error");
		return false;
	}

	if (_fullscreen)
	{
		DXGI_MODE_DESC md;
		DXUTGetD3D10AdapterDisplayMode(ds.d3d10.AdapterOrdinal, 0, &md);
		md.Width = _width;
		md.Height = _height;
		DXUTToggleFullScreen();
		DXUTGetDXGISwapChain()->ResizeTarget(&md);
	}

	_init = true;

	return true;
}

bool DXUTApp::Run()
{
	HRESULT hr = S_OK;	

	if ((hr = DXUTMainLoop()) != S_OK)
		PrintErrorMsg(logger, GetDXUTErrorMsg(hr), GetDXUTErrorMsg(hr), L"Runtime Error");

	return hr == S_OK;
}

int DXUTApp::Exit()
{
	return DXUTGetExitCode();
}

bool DXUTApp::LoadConfigs(const std::string& configFile)
{
	int fullscreen = 0;

	if (!FileIO::LoadValue(configFile, "width", _width) ||
		!FileIO::LoadValue(configFile, "height", _height) ||
		!FileIO::LoadValue(configFile, "fullscreen", fullscreen) ||
		!FileIO::LoadValue(configFile, "fov", _config_fov) ||
		!FileIO::LoadValue(configFile, "zNear", _config_zNear) ||
		!FileIO::LoadValue(configFile, "zFar", _config_zFar))
	{
		PrintErrorMsg(logger, L"Error loading configfile.", L"Error loading configfile.", L"Config Error");
		return false;
	}
	
	_fullscreen = fullscreen == 1;
	
	return true;
}

void DXUTApp::Update(float)
{
}

void DXUTApp::Display(float)
{
}

HRESULT DXUTApp::OnCreateDevice(ID3D10Device* d3dDevice, const DXGI_SURFACE_DESC*)
{
	HRESULT hr = S_OK;

    V_RETURN(_dialogResourceManager.OnD3D10CreateDevice(d3dDevice));
	V_RETURN(_settingsDlg.OnD3D10CreateDevice(d3dDevice));

	if (_init)
		if (!Initialize())
			hr = S_FALSE;

	return hr;
}

HRESULT DXUTApp::OnResize(ID3D10Device* d3dDevice, IDXGISwapChain* swapchain, const DXGI_SURFACE_DESC* bufferSurfaceDesc)
{
	HRESULT hr = S_OK;

	_swapChain = swapchain;

	V_RETURN(_dialogResourceManager.OnD3D10ResizedSwapChain(d3dDevice, bufferSurfaceDesc));
	V_RETURN(_settingsDlg.OnD3D10ResizedSwapChain(d3dDevice, bufferSurfaceDesc));

	return hr;
}

void DXUTApp::OnDestroyDevice()
{
	_dialogResourceManager.OnD3D10DestroyDevice();
	_settingsDlg.OnD3D10DestroyDevice();
	DXUTGetGlobalResourceCache().OnDestroyDevice();
}

void DXUTApp::OnReleasingSwapChain()
{
	_dialogResourceManager.OnD3D10ReleasingSwapChain();
}

bool DXUTApp::AllocateConsole()
{
	AllocConsole();

	*stdout = *_fdopen(_open_osfhandle(reinterpret_cast<long>(GetStdHandle(STD_OUTPUT_HANDLE)), _O_TEXT), "w" );

	setvbuf(stdout, NULL, _IONBF, 0 );

	std::ios::sync_with_stdio();

	return true;
}

bool DXUTApp::DeallocateConsole()
{
	return FreeConsole() != FALSE;
}

const std::wstring DXUTApp::GetDXUTErrorMsg(HRESULT hr) const
{
	std::wstring errorMsg(L"");

	switch (hr)
	{
		case DXUTERR_NODIRECT3D:
			errorMsg.assign(L"Could not initialize Direct3D.");
			break;
		case DXUTERR_NOCOMPATIBLEDEVICES:
			errorMsg.assign(L"Could not find any compatible Direct3D devices.");
			break;
		case DXUTERR_MEDIANOTFOUND:
			errorMsg.assign(L"Could not find required media.");
			break;
		case DXUTERR_NONZEROREFCOUNT:
			errorMsg.assign(L"The Direct3D device has a non-zero reference count, meaning some objects were not released.");
			break;
		case DXUTERR_CREATINGDEVICE:
			errorMsg.assign(L"An error occurred when attempting to create an Direct3D device.");
			break;
		case DXUTERR_RESETTINGDEVICE:
			errorMsg.assign(L"An error occurred when attempting to reset an Direct3D device.");
			break;
		case DXUTERR_CREATINGDEVICEOBJECTS:
			errorMsg.assign(L"An error occurred in the device create callback function.");
			break;
		case DXUTERR_RESETTINGDEVICEOBJECTS:
			errorMsg.assign(L"An error occurred in the device reset callback function.");
			break;
		case DXUTERR_DEVICEREMOVED:
			errorMsg.assign(L"The Direct3D device was removed.");
			break;
		default:
			errorMsg.assign(L"Unknown Error.");
			break;
	}
	return errorMsg;
}

void DXUTApp::PrintErrorMsg(Logger& logger, const std::wstring& loggerMsg, const std::wstring& msgBoxText, const std::wstring& msgBoxCaption) const
{
	if (logger.IsEnabled(Logger::Error))
		Logger::Stream(logger, Logger::Error) << Util::WStrToStr(loggerMsg).c_str();
		
	MessageBox(DXUTGetHWND(), msgBoxText.c_str() , msgBoxCaption.c_str(), MB_OK | MB_ICONERROR);	
}

void DXUTApp::EditSettingDlg()
{
	const WCHAR* types[4] =
	{
		L"D3D10_DRIVER_TYPE_REFERENCE",
		L"D3D10_DRIVER_TYPE_NULL",
		L"D3D10_DRIVER_TYPE_SOFTWARE",
		L"D3D10_DRIVER_TYPE_WARP",
	};

	CDXUTDialog* dialog = _settingsDlg.GetDialogControl();

	CDXUTComboBox* comboBox = dialog->GetComboBox(DXUTSETTINGSDLG_DEVICE_TYPE);
	// remove all device types except hardware device
	for (unsigned int i = 0; i < 4; ++i)
	{
		int index = comboBox->FindItem(types[i]);
		if (index >= 0)
			comboBox->RemoveItem(index);
	}
}

tVec2i DXUTApp::GetMousePos()
{
	POINT p;

	GetCursorPos(&p);
	ScreenToClient(DXUTGetHWND(), &p);

	return tVec2i(p.x, p.y);
}